# -*- coding: utf-8 -*-

import operator
import statistics
from collections import defaultdict
from copy import copy
from copy import deepcopy
from urllib.parse import unquote

from .exceptions import RQLQueryError
from .parser import Parser


class NodeMeta(type):
    nodes = {}

    def __init__(cls, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        if not (name.endswith("Node") or name.startswith("_")):
            NodeMeta.nodes[getattr(cls, "name", name.lower())] = cls


class Node(metaclass=NodeMeta):
    def __init__(self, *args):
        self.args = args

    @classmethod
    def get_subnode(cls, name):
        try:
            return NodeMeta.nodes[name]
        except KeyError:
            raise RQLQueryError("Invalid query function: %s" % name)


class RowNode(Node):
    def feed(self, data):
        return [row for row in data if self(row)]


class DataNode(Node):
    def feed(self, data):
        return self(data)


class Key(Node):
    def __init__(self, *args):
        self.args = []
        for arg in args:
            self.args.extend(arg.split("."))

    def __call__(self, row):
        value = row
        for key in self.args:
            value = value.get(key)

        return value

    def __str__(self):
        return ".".join(self.args)


class _Filter(RowNode):
    name = None

    def __init__(self, key, value):
        self.key = Key(key)
        self.value = value
        self.op = getattr(operator, self.name)

    def __call__(self, row):
        if isinstance(self.value, Key):
            value = self.value(row)
        else:
            value = self.value

        return self.op(self.key(row), value)


class EqualTo(_Filter):
    name = "eq"


class NotEqualTo(_Filter):
    name = "ne"


class LessThan(_Filter):
    name = "lt"


class LessOrEqual(_Filter):
    name = "le"


class GreaterThan(_Filter):
    name = "gt"


class GreaterOrEqual(_Filter):
    name = "ge"


class And(RowNode):
    def __call__(self, row):
        return all([f(row) for f in self.args if f is not None])


class Or(RowNode):
    def __call__(self, row):
        return any([f(row) for f in self.args if f is not None])


class In(RowNode):
    def __call__(self, row):
        key, value = self.args
        key = Key(key)
        return operator.contains(value, key(row))


class Out(RowNode):
    def __call__(self, row):
        key, value = self.args
        key = Key(key)
        return operator.not_(operator.contains(value, key(row)))


class Contains(RowNode):
    def __call__(self, row):
        key, value = self.args
        key = Key(key)
        return operator.contains(key(row), value)


class Excludes(RowNode):
    def __call__(self, row):
        key, value = self.args
        key = Key(key)
        return operator.not_(operator.contains(key(row), value))


class AggregateNode(DataNode):
    def __init__(self, key):
        self.key = Key(key)

    def __str__(self):
        return str(self.key)


class Min(AggregateNode):
    def __call__(self, data):
        return min([self.key(row) for row in data])


class Max(AggregateNode):
    def __call__(self, data):
        return max([self.key(row) for row in data])


class Sum(AggregateNode):
    def __call__(self, data):
        return sum([self.key(row) for row in data])


class Mean(AggregateNode):
    def __call__(self, data):
        return statistics.mean([self.key(row) for row in data])


class Select(DataNode):
    def __call__(self, data):
        keys = [Key(arg) for arg in self.args]
        return [{str(key): key(row) for key in keys} for row in data]


class Values(DataNode):
    def __init__(self, *args):
        if len(args) > 1:
            raise RQLQueryError("values() must have a single key argument")
        self.key = Key(args[0])

    def __call__(self, data):
        return [self.key(row) for row in data]


class Aggregate(DataNode):
    def __init__(self, key, *aggrs):
        self.key = Key(key)
        self.aggrs = aggrs

    def __call__(self, data):
        groups = defaultdict(list)
        for row in data:
            groups[self.key(row)].append(row)

        data = [
            {str(self.key): value, **{str(aggr): aggr(rows) for aggr in self.aggrs}}
            for (value, rows) in groups.items()
        ]
        return data


class Unwind(DataNode):
    def __init__(self, *args):
        if len(args) > 1:
            raise RQLQueryError("unwind() must have a single key argument")
        self.key = Key(args[0])

    def __call__(self, data):
        data = [{**row, str(self.key): item} for row in data for item in self.key(row)]
        return data


class Limit(DataNode):
    def __call__(self, data):
        limit, offset = self.args

        if offset:
            data = data[offset:]

        if limit and limit != float("inf"):
            data = data[:limit]

        return data


class Index(DataNode):
    def __call__(self, data):
        return data[self.args[0]]


class Slice(DataNode):
    def __call__(self, data):
        return data[slice(*self.args)]


class First(DataNode):
    def __call__(self, data):
        return data[:1]


class Count(DataNode):
    def __call__(self, data):
        return len(data)

    def __str__(self):
        return "count"


class Distinct(DataNode):
    def __call__(self, data):
        new_data = []

        for row in data:
            if row not in new_data:
                new_data.append(row)
        return new_data


class Sort(DataNode):
    def __init__(self, *args):
        self.args = [("+", v) if isinstance(v, str) else v for v in args]

    def __call__(self, data):
        # sort least significant first
        if self.args:
            for prefix, key in reversed(self.args):
                data.sort(key=operator.itemgetter(key), reverse=prefix == "-")
        else:
            data.sort()

        return data


class One(DataNode):
    def __call__(self, data):
        if len(data) > 1:
            raise RQLQueryError("Multiple results found for 'one'")

        if len(data) == 0:
            raise RQLQueryError("No results found for 'one'")

        return data


class Query:
    def __init__(self, data, default_limit=None, max_limit=None):
        self.data = data

        self._default_limit = default_limit
        self._max_limit = max_limit
        self._limit_clause = None

        self.rql_parsed = None
        self.rql_expr = ""

        self.pipeline = []

    def query(self, expr, ignore_top_eq=None):
        if not expr:
            return self

        new = copy(self)

        new.rql_expr = expr = unquote(expr)
        new.rql_parsed = Parser().parse(expr)

        # if there's a query, build the pipeline
        if new.rql_parsed:
            # if top-level node is not an 'and', make it so
            if new.rql_parsed["name"] != "and":
                new.rql_parsed = {"name": "and", "args": [new.rql_parsed]}

            # if we were asked to ignore any top level eq nodes,
            # remove them
            if ignore_top_eq:
                new.rql_parsed["args"] = [
                    arg
                    for arg in new.rql_parsed["args"]
                    if not (arg["name"] == "eq" and arg["args"][0] in ignore_top_eq)
                ]

            try:
                new.pipeline.extend(new._apply(new.rql_parsed).args)
            except RQLQueryError:
                raise
            except Exception as exc:
                raise RQLQueryError(f"{exc.__class__.__name__} preparing pipeline: {exc.args}")

        return new

    def all(self):
        # deepcopy data so we can transform it at will
        data = deepcopy(self.data)

        # execute the pipeline
        for node in self.pipeline:
            try:
                data = node.feed(data)
            except RQLQueryError:
                raise
            except Exception as exc:
                raise RQLQueryError(f"{exc.__class__.__name__} executing node {node}: {exc}")

        # if there's a default limit and no limit clause was added,
        # add one and feed the data through it
        if self._default_limit and self._limit_clause is None:
            data = Limit(self._default_limit, 0).feed(data)

        return data

    def _apply(self, token):
        if not isinstance(token, dict):
            return token

        name = token["name"]
        node_class = Node.get_subnode(name)

        args = [self._apply(arg) for arg in token["args"]]
        kwargs = {}

        # if node is Limit, apply limit constraints
        if node_class == Limit:
            limit = min(args[0] or float("inf"), self._max_limit or float("inf"))
            try:
                offset = args[1]
            except IndexError:
                offset = 0
            args = (limit, offset)

            self._limit_clause = True

        node = node_class(*args, **kwargs)

        return node



# -*- coding: utf-8 -*-

from datetime import datetime
from decimal import Decimal
from uuid import UUID

import pyparsing as pp
from dateutil.parser import parse as dateparse
from pyparsing import pyparsing_common as common
from six.moves import urllib

from .exceptions import RQLSyntaxError

# autoconvert:
# numbers
# booleans
# null

# converters:
# number
# epoch
# date
# datetime
# boolean
# string
# uuid
# decimal


def _sort_call(expr, loc, toks):
    return {"name": "sort", "args": toks.args.asList()}


def _array(expr, loc, toks):
    return tuple(toks)


def _unquote(expr, loc, toks):
    return urllib.parse.unquote(toks[0])


def _call_operator(expr, loc, toks):
    return toks.asDict()


def _comparison(expr, loc, toks):
    if len(toks) == 2:
        return {"name": "eq", "args": toks.asList()}

    else:
        op = toks.pop(1)
        return {"name": op, "args": toks.asList()}


def _or(expr, loc, toks):
    if len(toks) == 1:
        return toks[0]
    else:
        return {"name": "or", "args": toks.asList()}


def _and(expr, loc, toks):
    if len(toks) == 1:
        return toks[0]
    else:
        return {"name": "and", "args": toks.asList()}


def _group(expr, loc, toks):
    return toks[0]


def _date(expr, loc, toks):
    return dateparse(toks[0]).date()


def _datetime(expr, loc, toks):
    return dateparse(toks[0])


def _epoch(expr, loc, toks):
    return datetime.utcfromtimestamp(toks[0])


def _decimal(expr, loc, toks):
    return Decimal(toks[0])


def _uuid(expr, loc, toks):
    return UUID(hex=toks[0])


TRUE = pp.Keyword("true").setParseAction(pp.replaceWith(True))
FALSE = pp.Keyword("false").setParseAction(pp.replaceWith(False))
NULL = pp.Keyword("null").setParseAction(pp.replaceWith(None))

# let's treat sort as a keyword to better handle the +- prefix syntax
SORT = pp.Keyword("sort").suppress()

# keywords for typed values
K_NUMBER = pp.Keyword("number").suppress()
K_STRING = pp.Keyword("string").suppress()
K_DATE = pp.Keyword("date").suppress()
K_DATETIME = pp.Keyword("datetime").suppress()
K_BOOL = pp.Keyword("boolean").suppress()
K_EPOCH = pp.Keyword("epoch").suppress()
K_UUID = pp.Keyword("uuid").suppress()
K_DECIMAL = pp.Keyword("decimal").suppress()

# grammar
PLUS = pp.Literal("+")
MINUS = pp.Literal("-")
EQUALS = pp.Literal("=").suppress()
LPAR = pp.Literal("(").suppress()
RPAR = pp.Literal(")").suppress()
COLON = pp.Literal(":").suppress()

# reserved characters that are not part of the RQL grammar
RESERVED = pp.Word("@!*+$", exact=1)

UNRESERVED = pp.Word(pp.pyparsing_unicode.alphanums + "-:._~ ", exact=1)
PCT_ENCODED = pp.Combine(pp.Literal("%") + pp.Word(pp.hexnums, exact=2)).setParseAction(_unquote)
NCHAR = UNRESERVED | PCT_ENCODED | RESERVED

STRING = pp.Combine(pp.OneOrMore(NCHAR))

NAME = common.identifier

NUMBER = common.number

TYPED_STRING = K_STRING + COLON + STRING
TYPED_NUMBER = K_NUMBER + COLON + common.number
TYPED_DATE = (K_DATE + COLON + common.iso8601_date).setParseAction(_date)
TYPED_DATETIME = (K_DATETIME + COLON + common.iso8601_datetime).setParseAction(_datetime)
TYPED_BOOL = K_BOOL + COLON + (TRUE | FALSE)
TYPED_EPOCH = (K_EPOCH + COLON + common.number).setParseAction(_epoch)
TYPED_UUID = (K_UUID + COLON + STRING).setParseAction(_uuid)
TYPED_DECIMAL = (K_DECIMAL + COLON + STRING).setParseAction(_decimal)

TYPED_VALUE = (
    TYPED_DECIMAL | TYPED_UUID | TYPED_EPOCH | TYPED_DATETIME | TYPED_DATE | TYPED_NUMBER | TYPED_BOOL | TYPED_STRING
)

ARRAY = pp.Forward()

# using ^ instead of | between NUMBER and STRING to avoid ambiguity
# when parsing strings starting with numbers
VALUE = TYPED_VALUE | ARRAY | TRUE | FALSE | NULL | (NUMBER ^ STRING)

PAR_ARRAY = (LPAR + pp.delimitedList(VALUE) + RPAR).setParseAction(_array)

ARRAY <<= PAR_ARRAY

CALL_OPERATOR = pp.Forward()

ARGUMENT = CALL_OPERATOR | VALUE

SORT_ARG = ((MINUS | PLUS) + VALUE).setParseAction(lambda e, l, t: tuple(t))
SORT_ARGARRAY = pp.delimitedList(SORT_ARG).setResultsName("args")
SORT_CALL = (SORT + LPAR + SORT_ARGARRAY + RPAR).setParseAction(_sort_call)

FUNC_CALL = (
    NAME.setResultsName("name")
    + LPAR
    + pp.Group(pp.Optional(pp.delimitedList(ARGUMENT))).setResultsName("args")
    + RPAR
).setParseAction(_call_operator)

CALL_OPERATOR <<= SORT_CALL | FUNC_CALL

COMPARISON = (VALUE + EQUALS + pp.Optional(NAME + EQUALS) + VALUE).setParseAction(_comparison)

OPERATOR = pp.Forward()

OR = pp.delimitedList(OPERATOR, delim=pp.Literal("|")).setParseAction(_or)
AND = pp.delimitedList(OPERATOR, delim=pp.Literal("&")).setParseAction(_and)


GROUP = (LPAR + (OR | AND) + RPAR).setParseAction(_group)

OPERATOR <<= GROUP | COMPARISON | CALL_OPERATOR

QUERY = pp.delimitedList(AND).setParseAction(_and)


class Parser:
    def parse(self, expr):
        try:
            result = QUERY.parseString(expr, parseAll=True)
        except pp.ParseException as exc:
            raise RQLSyntaxError(*exc.args)

        return result[0]



# -*- coding: utf-8 -*-

import datetime
import decimal
import uuid


class Unparser:
    def unparse(self, expr):
        name = expr.get("name")

        args = []

        for a in expr.get("args", []):
            if isinstance(a, dict):
                arg = self.unparse(a)

            elif isinstance(a, tuple):
                arg = self.unparse_tuple(a)

            else:
                arg = self.unparse_token(a)

            args.append(arg)

        return "{}({})".format(name, ",".join(map(str, args)))

    def unparse_tuple(self, arg):
        prefix = ""
        tokens = []

        if arg[0] in {"+", "-"}:
            prefix = arg[0]
            arg = arg[1:]

        for a in arg:
            if isinstance(a, tuple):
                tokens.append(self.unparse_tuple(a))
            else:
                tokens.append(self.unparse_token(a))

        return prefix + "(" + ",".join(tokens) + ")"

    def unparse_token(self, arg):
        if arg is None:
            return "null"

        elif isinstance(arg, bool):
            return str(arg).lower()

        elif isinstance(arg, decimal.Decimal):
            return "decimal:%s" % arg

        elif isinstance(arg, float):
            # repr(float) returns the shortest decimal representation
            # for the same binary float
            return repr(arg)

        elif isinstance(arg, uuid.UUID):
            return "uuid:%s" % arg.hex

        elif isinstance(arg, datetime.datetime):
            return "datetime:%s" % arg.isoformat()

        elif isinstance(arg, datetime.date):
            return "date:%s" % arg.isoformat()

        return str(arg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrql']

package_data = \
{'': ['*']}

install_requires = \
['pyparsing>=2.4,<3.0', 'python-dateutil>=2.8,<3.0']

setup_kwargs = {
    'name': 'pyrql',
    'version': '0.7.5',
    'description': 'RQL parsing',
    'long_description': None,
    'author': 'Pedro Werneck',
    'author_email': 'pjwerneck@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pjwerneck/pyrql',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
